# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Literal
from typing import TypeVar

import pydantic

import aiopki
from aiopki.types import IAlgorithm
from aiopki.types import IDecrypter
from aiopki.types import IEncrypter
from aiopki.types import ISigner
from aiopki.types import IVerifier
from aiopki.utils import b64decode
from .types import JWE
from .types import JWS
from .types import JWT
from .types import Payload


T = TypeVar('T')


class JOSEObject(pydantic.RootModel[JWS | JWE | Payload]):

    @property
    def typ(self) -> str | None:
        return str.lower(self.root.typ) if self.root.typ else None

    @staticmethod
    def deserialize(value: str | bytes) -> dict[str, Any]:
        if isinstance(value, bytes):
            value = bytes.decode(value, 'ascii')
        params: dict[str, Any] = {}
        if value.count('.') == 2:
            protected, payload, signature = str.split(value, '.')
            params.update({
                'payload': payload,
                'signatures': [{
                    'protected': protected,
                    'signature': b64decode(signature)
                }]
            })
        if value.count('.') == 4:
            protected, key, iv, ct, tag = str.split(value, '.')
            
            params.update({
                'aad': str.encode(protected, 'ascii'),
                'ciphertext': ct,
                'iv': iv,
                'mode': 'compact',
                'protected': protected,
                'recipients': [{'encrypted_key': key}],
                'tag': tag,
            })
        return params

    @classmethod
    def jwe(
        cls,
        payload: JWS | dict[str, Any] | bytes,
        enc: str | IAlgorithm, 
        alg: str | IAlgorithm | None = None,
        cty: str | None = None,
        protected: dict[str, Any] | None = None,
        mode: Literal['compact', 'json', 'unset'] = 'unset',
    ) -> JWE:
        """Create a new JSON Web Encryption (JWE) instance.

        Args:
            payload: the payload of the JWE. Can be either a byte-sequences
                or dictionary.
            enc (:class:`aiopki.types.IAlgorithm`): specifies the algorithm that is
                used to encrypt content using the CEK.
            alg (:class:`aiopki.types.IAlgorithm`): specifies the algorithm that is
                used to encrypt the Content Encryption Key (CEK). If the `alg`
                parameter is provided, then all recipients are required to use this
                algorithm; otherwise each recipient must specify its own.
                See also :meth:`aiopki.ext.jose.types.JWE.add_recipient()`
            cty (str): optionally specifies the content type (``cty`` header),
                if it can not be inferred from the payload type.

        Returns:
            :class:`aiopki.ext.jose.JOSEObject`
        """
        protected = protected or {}
        if isinstance(alg, str):
            alg = aiopki.get(alg)
        if isinstance(enc, str):
            enc = aiopki.get(enc)
        if alg is not None:
            protected['alg'] = alg.name
        protected['enc'] = enc.name
        if cty:
            protected['cty'] = cty
        if isinstance(payload, JWS):
            payload, protected['cty'] = payload.serialize()
        obj = cls.model_validate({
            'ciphertext': b'',
            'mode': mode,
            'payload': payload,
            'protected': protected or b'',
            'recipients': [],
            'mode': mode
        })
        assert isinstance(obj.root, JWE)
        return obj.root

    @classmethod
    def parse_compact(cls, value: str | bytes):
        return cls.model_validate(cls.deserialize(value))

    @pydantic.model_validator(mode='before')
    def preprocess(cls, value: Any) -> Any:
        if isinstance(value, str):
            value = cls.deserialize(value)
        return value

    def payload(
        self,
        factory: Callable[[bytes], T] = JWT.model_validate
    ) -> T:
        assert not isinstance(self.root, Payload)
        obj = self.root
        if isinstance(obj, JWE):
            raise NotImplementedError
        assert obj.payload is not None
        return factory(b64decode(obj.payload))

    def encode(self, encoder: Callable[[bytes], T] = bytes, compact: bool = False) -> T:
        """Encodes the JWS/JWE.

        Args:
            encoder: encodes the result, if an encoding other than bytes
                is needed.
            compact (bool): indicates if compact encoding is required.
                Using ``compact=True`` with a JOSE object that has
                multiple recipients or signers, raises an :exc:`ValueError`.

        Returns:
            bytes
        """
        assert not isinstance(self.root, Payload)
        return encoder(self.root.encode(compact=compact))

    def is_encrypted(self) -> bool:
        return isinstance(self.root, JWE)

    async def add_recipient(
        self,
        encrypter: IEncrypter,
        enc: IAlgorithm | str = 'A128GCM',
        alg: IAlgorithm | str | None = None,
    ):
        if not isinstance(self.root, JWE):
            raise TypeError(f"Can not add recipients to {type(self.root).__name__}")
        if isinstance(alg, str):
            alg = aiopki.get(alg)
        if isinstance(enc, str):
            enc = aiopki.get(enc)
        return await self.root.add_recipient(
            encrypter=encrypter,
            alg=alg,
            enc=enc
        )

    async def decrypt(self, decrypter: IDecrypter) -> None:
        assert isinstance(self.root, JWE)
        self.root = await self.root.decrypt(decrypter)

        assert isinstance(self.root, Payload)
        if self.root.is_jose():
            assert self.root.is_compact()
            self.root = self.parse_compact(bytes(self.root)).root

    async def encrypt(
        self,
        encrypter: IEncrypter | None = None,
        alg: IAlgorithm | None = None,
        enc: IAlgorithm | None = None,
    ) -> None:
        """Encrypt the :class:`JOSEObject` using the appropriate means.

        If the object is JSON Web Encryption (JWE), use the Content Encryption
        Key (CEK) to encrypt the payload if Key Encryption, Key Agreement or
        Key Wrapping is used.

        Args:
            encrypter (:class:`aiopki.types.IEncrypter`): the encrypter used to
                encrypt the payload, or ``None`` if the payload knows how to
                encrypt itself.

        Returns:
            None

        Raises:
            :exc:`TypeError`: the `encrypter` parameter is ``None`` and the
                payload can not encrypt itself.
            :exc:`TypeError`: the object is a JWE with multiple recipients
                and `compact` is ``True``.
        """
        if isinstance(self.root, JWE):
            pass
        elif isinstance(self.root, JWS):
            if encrypter is None:
                raise TypeError("The `encrypter` argument is required.")
            self.root = self.jwe(
                payload=self.root,
                alg=alg,
                enc=enc or 'A256GCM'
            )
            await self.root.add_recipient(encrypter)
        else:
            raise NotImplementedError

    async def sign(
        self,
        algorithm: str | IAlgorithm,
        signer: ISigner,
        protected: dict[str, Any] | None = None,
        header: dict[str, Any] | None = None
    ) -> None:
        """Add a signature to a JWS object."""
        if not isinstance(self.root, JWS):
            raise NotImplementedError
        if isinstance(algorithm, str):
            algorithm = aiopki.get(algorithm)
        await self.root.sign(algorithm, signer, protected or {}, header or {})

    async def verify(self, verifier: IVerifier) -> bool:
        """Return a boolean indicating if at least one signature
        validated using the given verifier. Raise an exception if
        the object does not have signatures (i.e. is JWE).
        """
        if not isinstance(self.root, JWS):
            raise NotImplementedError
        return await self.root.verify(verifier)

    def __bytes__(self) -> bytes:
        if not isinstance(self.root, Payload):
            raise TypeError(f'Can not cast {type(self).__name__} to bytes')
        return bytes(self.root)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey

from aiopki.tests.compat.jwe import *


@pytest.fixture(scope='module')
def alg() -> str:
    return 'RSA-OAEP'


@pytest.fixture(scope='module')
def enc() -> str:
    return 'A192GCM'


@pytest.fixture(scope='module')
def encrypter(decrypter: JSONWebKey) -> JSONWebKey:
    assert decrypter.public is not None
    return decrypter.public


@pytest.fixture(scope='module')
def decrypter(alg: str, rsa_key: RSAPrivateKey) -> JSONWebKey:
    return JSONWebKey.model_validate({
        'kty': 'RSA',
        'use': 'enc',
        'alg': alg,
        'key': rsa_key,
        'key_ops': ['wrapKey', 'unwrapKey']
    })



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .base64 import Base64
from .base64json import Base64JSON
from .cryptokeyspecification import CryptoKeySpecification
from .encryptionresult import EncryptionResult
from .ialgorithm import IAlgorithm
from .icryptobackend import ICryptoBackend
from .icryptokey import ICryptoKey
from .idecrypter import IDecrypter
from .iencrypter import IEncrypter
from .isecret import ISecret
from .isigner import ISigner
from .itruststore import ITrustStore
from .iverifier import IVerifier
from .jsonobject import JSONObject
from .plaintext import Plaintext
from .subjectid import SubjectID
from .undecryptable import Undecryptable

__all__: list[str] = [
    'Base64',
    'Base64JSON',
    'CryptoKeySpecification',
    'EncryptionResult',
    'IAlgorithm',
    'ICryptoBackend',
    'ICryptoKey',
    'IDecrypter',
    'IEncrypter',
    'ISecret',
    'ISigner',
    'ITrustStore',
    'IVerifier',
    'JSONObject',
    'Plaintext',
    'SubjectID',
    'Undecryptable',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Union

import pydantic

from aiopki.algorithms import get as get_algorithm
from aiopki.types import EncryptionResult
from aiopki.types import IAlgorithm
from aiopki.types import Plaintext
from .edwardscurveprivatekey import EdwardsCurvePrivateKey
from .edwardscurvepublickey import EdwardsCurvePublicKey
from .ellipticcurveprivatekey import EllipticCurvePrivateKey
from .ellipticcurvepublickey import EllipticCurvePublicKey
from .octetsigningkey import OctetSigningKey
from .rsaprivatekey import RSAPrivateKey
from .rsapublickey import RSAPublicKey
from .symmetricencryptionkey import SymmetricEncryptionKey
from .symmetricwrappingkey import SymmetricWrappingKey


__all__: list[str] = ['JSONWebKey']

JWKType = Union[
    EdwardsCurvePrivateKey,
    EllipticCurvePrivateKey,
    SymmetricEncryptionKey,
    SymmetricWrappingKey,
    OctetSigningKey,
    RSAPrivateKey,
    EdwardsCurvePublicKey,
    EllipticCurvePublicKey,
    RSAPublicKey,
]


class JSONWebKey(pydantic.RootModel[JWKType]):

    @property
    def alg(self) -> str | None:
        return self.root.alg

    @property
    def crv(self) -> str | None:
        return self.root.crv

    @property
    def kid(self) -> str | None:
        return self.root.kid

    @property
    def key_ops(self) -> tuple[str, ...]:
        return tuple(self.root.key_ops or [])

    @property
    def kty(self) -> str:
        return self.root.kty

    @property
    def public(self) -> Union['JSONWebKey', None]:
        if self.root.public:
            return JSONWebKey.model_validate(self.root.public.model_dump())

    @property
    def thumbprint(self) -> str:
        return self.root.thumbprint

    @property
    def use(self) -> str | None:
        return self.root.use

    def can_decrypt(self) -> bool:
        return self.root.can_decrypt()

    def can_encrypt(self) -> bool:
        return self.root.can_encrypt()

    def can_use(self, algorithm: IAlgorithm) -> bool:
        return self.root.can_use(algorithm)

    def default_algorithm(self) -> IAlgorithm:
        if self.root.alg is not None:
            return get_algorithm(self.root.alg)
        if not self.root.use:
            raise ValueError(
                "Can not determine default algorithm if the `use` "
                "claim is not specified."
            )
        return self.root.default_algorithm()

    def encrypt_sync(
        self,
        pt: bytes | Plaintext,
        algorithm: IAlgorithm | None = None
    ) -> EncryptionResult:
        return self.root.encrypt_sync(pt, algorithm)

    def is_active(self) -> bool:
        if self.root.nbf is None:
            return True
        now = datetime.datetime.now(datetime.timezone.utc).timestamp()
        return self.root.nbf <= now

    def is_available(self) -> bool:
        return not self.is_expired() and self.is_active()

    def is_expired(self) -> bool:
        if self.root.exp is None:
            return False
        now = datetime.datetime.now(datetime.timezone.utc).timestamp()
        return now >= self.root.exp

    def is_public(self) -> bool:
        return isinstance(self.root, (
            EdwardsCurvePublicKey,
            EllipticCurvePublicKey,
            RSAPublicKey,
        ))

    def process_signature(self, signature: bytes) -> bytes:
        return self.root.process_signature(signature)

    async def decrypt(
        self,
        ct: EncryptionResult,
        algorithm: IAlgorithm | None = None
    ) -> Plaintext:
        return await self.root.decrypt(ct, algorithm)

    async def encrypt(
        self,
        pt: bytes | Plaintext,
        algorithm: IAlgorithm | None = None
    ) -> EncryptionResult:
        return await self.root.encrypt(pt, algorithm)

    async def sign(
        self,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bytes:
        if algorithm is None:
            algorithm = self.root.get_algorithm()
        return await self.root.sign(message, algorithm) # type: ignore

    async def verify(
        self,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bool:
        if not self.is_available():
            return False
        if using is not None and using not in {self.kid, self.thumbprint}:
            return False
        if algorithm is None:
            algorithm = self.root.get_algorithm()
        return await self.root.verify(signature, message, algorithm, using) # type: ignore

    def __bytes__(self) -> bytes:
        return bytes(self.root)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import aiopki
from aiopki import CryptoKeyVersion
from aiopki.types import EncryptionResult, IAlgorithm, Plaintext


class LocalKey(aiopki.VersionedCryptoKey):

    async def discover(self) -> None:
        for k in self.versions:
            self._index[k.thumbprint] = k

    async def decrypt(
        self,
        version: CryptoKeyVersion,
        ct: EncryptionResult,
        algorithm: IAlgorithm
    ) -> Plaintext:
        assert version.key is not None
        return await version.key.decrypt(ct, algorithm)

    async def encrypt(
        self,
        version: CryptoKeyVersion,
        pt: Plaintext,
        algorithm: IAlgorithm
    ) -> EncryptionResult:
        assert version.key is not None
        return await version.key.encrypt(pt, algorithm)

    async def sign(
        self,
        version: CryptoKeyVersion,
        message: bytes,
        algorithm: IAlgorithm
    ) -> bytes:
        assert version.key is not None
        return await version.key.sign(message, algorithm)

    async def verify(
        self,
        version: CryptoKeyVersion,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm
    ) -> bool:
        assert version.key is not None
        return await version.key.verify(signature, message, algorithm)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

import aiopki


@pytest.mark.asyncio
async def test_encrypt_decrypt_bytes(
    encryption_key: aiopki.CryptoKeyType
):
    ct = await encryption_key.encrypt(b'Hello world!')
    pt = await encryption_key.decrypt(ct)
    assert bytes(pt) == b'Hello world!'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Protocol
from typing import TypeVar

from cryptography.hazmat.primitives.ciphers.modes import GCM

from .encryptionresult import EncryptionResult


T = TypeVar('T')


class IAlgorithm(Protocol):
    __module__: str = 'aiopki.types'
    name: str

    def can_encrypt(self) -> bool: ...
    def can_sign(self) -> bool: ...
    def cek(self, factory: Callable[[dict[str, Any]], T]) -> T: ...
    def get_digest_algorithm(self) -> Any: ...
    def get_padding(self) -> Any: ...

    def digest(self, value: bytes) -> bytes:
        raise NotImplementedError

    def get_curve_name(self) -> str | None:
        raise NotImplementedError

    def get_digest_name(self) -> str:
        raise NotImplementedError

    def get_initialization_vector(self, iv: bytes | None = None, tag: bytes | None = None) -> tuple[GCM, bytes]:
        raise NotImplementedError

    def is_direct(self) -> bool:
        """Return a boolean indicating if the algorithm is the ``dir``
        algorithm as defined in :rfc:`7518`.
        """
        return False

    def length(self) -> int:
        raise NotImplementedError

    def supports_aad(self) -> bool:
        raise NotImplementedError

    async def unwrap(
        self,
        algorithm: 'IAlgorithm',
        decrypter: Any,
        ct: EncryptionResult,
        factory: Callable[[dict[str, Any]], T]
    ) -> T:
        raise NotImplementedError

    def __repr__(self) -> str:
        return f'<{type(self).__name__}: {self.name}>'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from urllib.parse import parse_qsl
from urllib.parse import urlunparse
from urllib.parse import ParseResult

from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.hazmat.primitives.serialization import load_pem_public_key

from aiopki.lib import JSONWebKey
from aiopki.types import CryptoKeySpecification
from .cryptokey import LocalKey


def parse(name: CryptoKeySpecification) -> LocalKey | None:
    if isinstance(name, dict):
        jwk = JSONWebKey.model_validate(name)
        if jwk.alg is None:
            raise ValueError("JWK input must specify the `alg` claim.")
        obj = LocalKey(
            backend='localhost',
            name=jwk.thumbprint,
            default_version=jwk.thumbprint
        )
        obj.add_version(
            name=jwk.thumbprint,
            alg=jwk.alg,
            thumbprint=jwk.thumbprint,
            public=jwk.public,
            key=jwk
        )
        return obj
    elif not isinstance(name, ParseResult) or name.scheme != 'file':
        return
    path = name.path[1:]
    if not path:
        raise ValueError(f"Invalid DSN: {urlunparse(name)}")
    with open(path, 'rb') as f:
        buf = f.read()
    is_pem = buf.startswith(b'-----')
    if not is_pem:
        raise NotImplementedError

    # TODO
    try:
        key = load_pem_public_key(buf)
    except ValueError:
        key = load_pem_private_key(buf, None)

    jwk = JSONWebKey.model_validate({
        **dict(parse_qsl(name.query)),
        'key': key
    })

    assert jwk.alg is not None

    obj = LocalKey(
        backend='localhost',
        name=path,
        default_version=jwk.thumbprint
    )
    obj.add_version(
        name=jwk.thumbprint,
        alg=jwk.alg,
        thumbprint=jwk.thumbprint,
        public=jwk.public,
        key=jwk
    )
    return obj


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest
from jwcrypto import jwk
from jwcrypto import jwe
from jwcrypto.common import json_encode # type: ignore

from aiopki.ext import jose
from aiopki.lib import JSONWebKey


@pytest.mark.asyncio
async def test_compact_decrypt_bytes(
    alg: str,
    enc: str,
    encrypter: JSONWebKey,
    decrypter: JSONWebKey,
    payload_bytes: bytes
) -> None:
    k = jwk.JWK(**encrypter.model_dump(exclude_none=True, mode='json')) # type: ignore
    protected = json_encode({
        'alg': alg,
        'enc': enc
    })
    o1 = jwe.JWE(payload_bytes, protected)
    o1.add_recipient(k) # type: ignore
    t = o1.serialize(compact=True)
    o2 = jose.parse(t)
    assert o2.is_encrypted()
    await o2.decrypt(decrypter)
    assert not o2.is_encrypted()
    assert bytes(o2) == payload_bytes


@pytest.mark.asyncio
async def test_compact_encrypt_bytes(
    alg: str,
    enc: str,
    encrypter: JSONWebKey,
    decrypter: JSONWebKey,
    payload_bytes: bytes
) -> None:
    k = jwk.JWK(**decrypter.model_dump(exclude_none=True, mode='json')) # type: ignore
    o1 = jose.jwe(payload_bytes, alg=alg, enc=enc)
    await o1.add_recipient(encrypter)
    t = o1.encode(encoder=bytes.decode, compact=True)
    o2 = jwe.JWE()
    o2.deserialize(t, k) # type: ignore
    assert o2.payload == payload_bytes # type: ignore


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .lib import JSONWebKeySet
from .types import EncryptionResult
from .types import IAlgorithm
from .types import ITrustStore
from .types import Plaintext
from .resource import Resource
from .versionedcryptokey import VersionedCryptoKey


__all__: list[str] = [
    'CryptoKeyType'
]


class CryptoKeyType(Resource[VersionedCryptoKey]):
    __module__: str = 'aiopki'
    model = VersionedCryptoKey

    @property
    def crv(self) -> str | None:
        return None

    @property
    def thumbprint(self) -> str:
        return self.impl.thumbprint

    def as_jwks(self) -> JSONWebKeySet:
        return self.impl.as_jwks()

    def can_decrypt(self) -> bool:
        return self.impl.can_decrypt()

    def can_use(self, algorithm: IAlgorithm) -> bool:
        return self.impl.can_use(algorithm)

    def default_algorithm(self) -> IAlgorithm:
        return self.impl.default_algorithm()

    def get_encryption_algorithms(self) -> set[str]:
        return {x.alg for x in self.impl.versions if x.can_encrypt()}

    def get_signing_algorithms(self) -> set[str]:
        return {x.alg for x in self.impl.versions if x.can_sign()}

    async def discover(self):
        if not self.ready:
            await self.impl
        return self

    async def decrypt(
        self,
        ct: EncryptionResult,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> Plaintext:
        version = self.impl.get(using)
        algorithm = algorithm or version.default_algorithm()
        if not version.can_use(algorithm):
            raise ValueError(f"Algorithm {algorithm} can not be used with this key.")
        return await self.impl.decrypt(
            version=version,
            ct=ct,
            algorithm=algorithm
        )

    async def encrypt(
        self,
        pt: bytes | Plaintext,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> EncryptionResult:
        version = self.impl.get(using)
        algorithm = algorithm or version.default_algorithm()
        if not version.can_use(algorithm):
            raise ValueError(f"Algorithm {algorithm} can not be used with this key.")
        if not isinstance(pt, Plaintext):
            pt = Plaintext(pt=pt)
        if version.is_asymmetric():
            return await version.encrypt(pt, algorithm)
        return await self.impl.encrypt(
            version=version,
            pt=pt,
            algorithm=algorithm
        )

    async def sign(
        self,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bytes:
        await self.impl
        version = self.impl.get(using)
        return await self.impl.sign(
            version=version,
            message=message,
            algorithm=algorithm or version.default_algorithm()
        )

    async def verify(
        self,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bool:
        await self.impl
        if using and not self.impl.has(using):
            return False
        version = self.impl.get(using)
        return await self.impl.verify(
            version=version,
            signature=signature,
            message=message,
            algorithm=algorithm or version.default_algorithm(),
        )

    async def trust(self, truststore: ITrustStore) -> None:
        for version in self.impl.versions:
            await truststore.trust(version)

    def __await__(self):
        return self.discover().__await__()

    def __or__(self, key: 'CryptoKeyType') -> JSONWebKeySet:
        s1 = self.as_jwks()
        s2 = key.as_jwks()
        return JSONWebKeySet(keys=[*s1.keys, *s2.keys])


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets
import os
from typing import Any
from typing import Callable
from typing import Literal
from typing import TypeVar

from cryptography.hazmat.primitives.ciphers.modes import CBC
from cryptography.hazmat.primitives.ciphers.modes import GCM

from aiopki.types import EncryptionResult
from aiopki.types import IAlgorithm
from aiopki.types import IDecrypter
from aiopki.utils import b64encode
from .encryptionalgorithm import EncryptionAlgorithm


T = TypeVar('T')


class AESEncryption(EncryptionAlgorithm):
    __module__: str = 'oauthx.algorithms'
    iv_length: int = 12
    _modes: dict[str, type[CBC | GCM]] = {
        'CBC': CBC,
        'GCM': GCM
    }

    def __init__(self, size: int, mode: Literal['CBC', 'GCM']) -> None:
        self._mode = mode
        self._size = size

    def can_encrypt(self) -> bool:
        return True

    def can_sign(self) -> bool:
        return False

    def cek(self, factory: Callable[[dict[str, Any]], T]) -> T:
        return factory({
            'kty': 'oct',
            'alg': self.name,
            'key_ops': ['wrapKey', 'unwrapKey'],
            'k': b64encode(secrets.token_bytes(self._size))
        })

    def get_initialization_vector(self, iv: bytes | None = None, tag: bytes | None = None) -> tuple[GCM, bytes]:
        Mode = self._modes[self._mode]
        iv = iv or os.urandom(self.iv_length)
        
        assert Mode == GCM
        return Mode(iv, tag=tag), iv # type: ignore

    def length(self) -> int:
        return self._size

    def supports_aad(self) -> bool:
        return self._mode == 'GCM'

    async def unwrap(
        self,
        algorithm: IAlgorithm,
        decrypter: IDecrypter,
        ct: EncryptionResult,
        factory: Callable[[dict[str, Any]], T]
    ) -> T:
        pt = await decrypter.decrypt(ct, algorithm)
        return factory({
            'kty': 'oct',
            'alg': 'dir',
            'key_ops': ['decrypt', 'encrypt'],
            'k': b64encode(bytes(pt))
        })


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from aiopki.lib import JSONWebKey
from aiopki.types import EncryptionResult
from aiopki.types import IAlgorithm
from aiopki.types import Plaintext
from .algorithms import get as algorithm


class CryptoKeyVersion(pydantic.BaseModel):
    name: str
    alg: str
    enabled: bool
    thumbprint: str
    public: JSONWebKey | None = None
    key: JSONWebKey | None = None

    @property
    def kid(self) -> str:
        t = self.thumbprint
        if self.public and self.public.kid:
            t = self.public.kid
        return t

    def can_decrypt(self) -> bool:
        return self.default_algorithm().can_encrypt()

    def can_encrypt(self) -> bool:
        return self.default_algorithm().can_encrypt()

    def can_sign(self) -> bool:
        return self.default_algorithm().can_sign()

    def can_use(self, algorithm: IAlgorithm) -> bool:
        return self.alg == algorithm.name

    def default_algorithm(self) -> IAlgorithm:
        return algorithm(self.alg)

    async def encrypt(
        self,
        pt: bytes | Plaintext,
        algorithm: IAlgorithm
    ) -> EncryptionResult:
        if self.public is None:
            raise NotImplementedError("Can only encrypt with asymmetric encryption.")
        return await self.public.encrypt(pt, algorithm=algorithm)

    def is_asymmetric(self) -> bool:
        return self.public is not None

    def process_signature(self, signature: bytes) -> bytes:
        if self.public is not None:
            signature = self.public.process_signature(signature)
        return signature



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
from typing import Any

from aiopki import BaseCryptoKey
from aiopki import CryptoKeyType
from aiopki import CryptoKeyVersion
from aiopki.lib import JSONWebKey
from aiopki.lib import JSONWebKeySet
from aiopki.types import IAlgorithm
from aiopki.types import ICryptoKey


__all__: list[str] = [
    'Keychain'
]


class Keychain(BaseCryptoKey):
    keys: list[CryptoKeyType] = []
    _default: str | None = None
    _discovered: bool = False
    _index: dict[str, CryptoKeyType] = {}
    _trust: dict[str, CryptoKeyVersion] = {}

    @property
    def crv(self) -> str:
        # TODO: remove this attribute
        return self.default().crv # type: ignore

    def as_jwks(self) -> JSONWebKeySet:
        keys: list[JSONWebKey] = []
        for _, obj in self._trust.items():
            if not obj.is_asymmetric() or obj.public is None:
                continue
            keys.append(obj.public)
        return JSONWebKeySet.model_validate({'keys': keys})

    def model_post_init(self, _: Any) -> None:
        self._index = {}
        self._trust = {}

    def can_use(self, algorithm: IAlgorithm) -> bool:
        return True

    def default(self) -> ICryptoKey:
        assert self._default
        return self._index[self._default] # type: ignore

    def default_algorithm(self) -> IAlgorithm:
        return self.default().default_algorithm()

    def get(self, using: str) -> ICryptoKey:
        return self._index[using] # type: ignore

    def get_thumbprint(self) -> str:
        return self.default().get_thumbprint()

    async def discover(self):
        if not self._discovered:
            await asyncio.gather(*map(asyncio.ensure_future, self.keys))
            self._default = self.keys[0].thumbprint
            for key in self.keys:
                self._index[key.thumbprint] = key
                await key.trust(self)
        self._discovered = True
        return self

    async def trust(self, key: CryptoKeyVersion) -> None:
        if key.public is None:
            return
        self._trust[key.thumbprint] = key

    async def sign(
        self,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bytes:
        key = self.get(using) if using else self.default()
        return await key.sign(message, algorithm)

    async def verify(
        self,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm | None = None,
        using: str | None = None
    ) -> bool:
        if using not in self._trust:
            raise NotImplementedError(using, list(self._trust))
        k = self._trust[using]
        assert k.public
        return await k.public.verify(signature, message, algorithm)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Literal
from typing import TypeVar

import pydantic

from aiopki.algorithms import get as get_algorithm
from aiopki.lib import JSONWebKey
from aiopki.types import Base64
from aiopki.types import Base64JSON
from aiopki.types import EncryptionResult
from aiopki.types import IAlgorithm
from aiopki.types import IDecrypter
from aiopki.types import IEncrypter
from aiopki.types import IVerifier
from aiopki.types import JSONObject
from aiopki.types import Plaintext
from aiopki.types import Undecryptable
from aiopki.utils import b64encode
from aiopki.utils import b64encode_json
from .jws import JWS
from .payload import Payload
from .recipient import Recipient


T = TypeVar('T')


class JWE(pydantic.BaseModel):
    protected: Base64JSON = pydantic.Field(
        default_factory=Base64JSON
    )

    unprotected: JSONObject = pydantic.Field(
        default_factory=JSONObject
    )

    ciphertext: Base64

    # TODO: differs based on compact or json encoding
    aad: Base64 = pydantic.Field(
        default_factory=Base64
    )

    mode: Literal['compact', 'json', 'unset'] = pydantic.Field(
        default=...
    )

    iv: Base64 = pydantic.Field(
        default_factory=Base64
    )
    tag: Base64 = pydantic.Field(
        default_factory=Base64
    )

    payload: Base64 | None = None

    recipients: list[Recipient] = pydantic.Field(
        default=...
    )

    _cek: JSONWebKey | None = pydantic.PrivateAttr(default=None)

    @property
    def alg(self) -> IAlgorithm | None:
        if self.claims.get('alg'):
            return get_algorithm(self.claims['alg'])

    @property
    def claims(self) -> dict[str, Any]:
        return {
            **(self.unprotected or {}),
            **(self.protected or {})
        }

    @property
    def cty(self) -> str | None:
        return self.claims.get('cty')

    @property
    def typ(self) -> str | None:
        return self.claims.get('typ')

    @property
    def enc(self) -> IAlgorithm:
        return get_algorithm(self.claims['enc'])

    def encode(
        self,
        compact: bool = False,
        encoder: Callable[[bytes], T] = bytes
    ) -> T:
        """Encodes the JWS/JWE."""
        if self.mode != 'compact' or not compact:
            raise NotImplementedError
        recipient, *recipients = self.recipients
        kwargs: dict[str, Any] = {
            'mode': 'json'
        }
        exclude = kwargs.setdefault('exclude', {
            'mode',
            'payload'
        })
        
        # If there is only one recipient, add its algorithm
        # to the protected header.
        if not recipients and recipient.header.get('alg'):
            self.protected['alg'] = recipient.header['alg']

        if self.mode == 'compact':
            # In Compact Serialization there are no recipients included
            # in the encoded result. There is also no unprotected header.
            assert not self.unprotected
            exclude.update({'unprotected', 'recipients'})

        if not self.ciphertext:
            self._encrypt()
        params = {k: v for k, v in self.model_dump(**kwargs).items() if v}
        params.setdefault('key', '')
    
        if self.recipients and self._cek:
            assert self.recipients[0].encrypted_key is not None
            params['key'] = b64encode(self.recipients[0].encrypted_key, encoder=bytes.decode)
        result = '{protected}.{key}.{iv}.{ciphertext}.{tag}'.format(**params)
        return encoder(str.encode(result, 'ascii'))

    def model_post_init(self, _: Any) -> None:
        protected = self.protected or {}
        unprotected = self.unprotected or {}
        if bool(set(protected) & set(unprotected)):
            raise ValueError(
                "The claims in the protected and unprotected "
                "header must be disjoint."
            )
        for recipient in self.recipients:
            recipient.update_header(self.claims)

        # If there are no recipients, then we are creating a new
        # JWE. Generate a Content Encryption Key (CEK). For each
        # recipient we will add this key. If Direct Encryption is
        # used, then the CEK is not used (because the recipient
        # key is the CEK).
        if self.alg and self.alg.name != 'dir':
            self._cek = self.enc.cek(factory=JSONWebKey.model_validate)

    def get_ciphertext(self, recipient: Recipient) -> EncryptionResult:
        return EncryptionResult.model_validate({
            'alg': recipient.alg.name,
            'ct': self.ciphertext,
            'aad': self.aad,
            'iv': recipient.iv or self.iv,
            'tag': self.tag
        })

    async def add_recipient(
        self,
        encrypter: IEncrypter,
        enc: IAlgorithm | None = None,
        alg: IAlgorithm | None = None
    ):
        alg = alg or encrypter.default_algorithm()
        enc = self.enc or get_algorithm('A128GCM')
        params: dict[str, Any] = {}
        if self.payload is None:
            raise ValueError("Cannot encrypt without a payload.")
        if alg and self.alg and self.alg.name != alg.name:
            raise TypeError(f'Algorithm {alg.name} can not be used with this JWE.')
        if enc and self.enc and self.enc.name != enc.name:
            raise TypeError(f'Algorithm {enc.name} can not be used with this JWE.')
        if alg.name == 'dir' and self.recipients:
            raise TypeError(
                "Multiple recipients are not allowed in Direct Encryption "
                "mode."
            )
        if alg.name == 'dir':
            self.protected['alg'] = alg.name
        elif not self._cek:
            self._cek = self.enc.cek(factory=JSONWebKey.model_validate)
        if not alg.is_direct():
            assert self._cek is not None
            ct = await encrypter.encrypt(
                pt=bytes(self._cek),
                algorithm=alg
            )
            assert not ct.aad
            assert not ct.iv
            assert not ct.tag
            params['header'] = {'alg': alg.name}
            params['encrypted_key'] = bytes(ct)

        self.recipients.append(Recipient.model_validate(params))
        if len(self.recipients) > 1:
            self.mode = 'json'

        # If Direct Encryption is used, then the encrypter is the CEK
        # and thus must the payload be encrypted.
        if alg.name == 'dir':
            await self._encrypt_direct(encrypter)
        self.mode = 'compact'

    async def decrypt(self, decrypter: IDecrypter) -> JWS | Payload:
        candidates: list[Recipient] = []
        for recipient in self.recipients:
            if not recipient.can_decrypt(decrypter):
                continue
            candidates.append(recipient)
        if not candidates:
            raise Undecryptable(
                "The provided decrypter is not usable with "
                "any of the recipients."
            )
        payload = None
        for recipient in candidates:
            ct = self.get_ciphertext(recipient)
            try:
                payload = await recipient.decrypt(decrypter, ct)
            except Undecryptable:
                continue
        if payload is None:
            raise Undecryptable
        return Payload(cty=self.cty, value=bytes(payload))

    async def verify(self, verifier: IVerifier) -> bool:
        raise NotImplementedError("JWE does not have a signature.")

    async def _encrypt_direct(self, cek: IEncrypter) -> None:
        assert self.payload is not None
        ct = await cek.encrypt(
            pt=Plaintext(
                pt=bytes(self.payload),
                aad=b64encode_json(self.protected)
            ),
            algorithm=None
        )
        self._update_cek_encryption_result(ct)

    def _encrypt(self) -> None:
        assert self._cek is not None
        assert self.payload is not None
        ct = self._cek.encrypt_sync(
            pt=Plaintext(
                pt=bytes(self.payload),
                aad=b64encode_json(self.protected)
            ),
            algorithm=None
        )
        self._update_cek_encryption_result(ct)

    def _update_cek_encryption_result(self, ct: EncryptionResult) -> None:
        assert ct.aad
        assert ct.iv
        assert ct.tag
        self.aad, self.ciphertext, self.iv, self.tag = (
            Base64(ct.aad),
            Base64(bytes(ct)),
            Base64(ct.iv),
            Base64(ct.tag)
        )